/**
 * ArrayList.java
 * 2/24/12
 */

import java.util.Random;

/**
 * Implements an expandable list of generic objects.
 * 
 * @author CPSC170A
 */
public class ArrayList<T> {
    
    /** Array that objects are stored in. */
    private T[] objects;

    /** Number of objects currently stored in this list. */
    private int size;

    /** Initial capacity of the object array. */
    private static final int INIT_CAPACITY = 5;

    /**
     * Constructor, initializes an empty list.
     */
    @SuppressWarnings("unchecked")
    public ArrayList() {
	objects = (T[]) new Object[INIT_CAPACITY];
	size = 0;
    }

    /**
     * Add the specified object to the end of this array list.
     *
     * @param object object to add
     */
    public final void add(final T object) {
	add(object, size);
    }

    /**
     * Add the specified object to this array list at the specified
     * index.  Objects in the list at the specified index and higher
     * are shifted over to make room for the new object.
     *
     * @param object object to add
     * @param index index to add the object
     * @throws IndexOutOfBoundsException if index < 0 or index > size
     */
    public final void add(final T object, final int index) {
	if (index < 0 || index > size) {
	    throw new IndexOutOfBoundsException("Invalid index: " + index);
	}
	if (size == objects.length) {
	    doubleCapacity();
	}
	for (int objIter = size; objIter > index; objIter--) {
	    objects[objIter] = objects[objIter - 1];
	}
	objects[index] = object;
        size++;
    }


    /**
     * Add the objects in the specified array list to the end of this
     * array list. The objects should be added in the order that they
     * are stored.
     * 
     * @param newObjects list of objects to add
     * @throws NullPointerException if newObjects is null
     */
    public final void addList(final ArrayList<T> newObjects) {
        addList(newObjects, size);
    }

    /**
     * Add the objects in the specified array list to the specified
     * location in this array list. The objects should be added in the
     * order that they are stored. Objects that at the specified index
     * and higher are shifted up by the number of objects that are
     * added in order to make room for the new objects. If the array
     * of objects is null or empty this array list is unchanged.
     * 
     * @param newObjects list of objects to add
     * @param index to add list of objects
     * @throws IndexOutOfBoundsException if index < 0 or index > size
     * @throws NullPointerException if newObjects is null
     */
    public final void addList(final ArrayList<T> newObjects, final int index) {
	if (index < 0 || index > size) {
	    throw new IllegalArgumentException("Invalid index: " + index);
	}
	if (newObjects == null) {
	    throw new NullPointerException("Invalid list:" + newObjects);
	}
	for (int i = 0; i < newObjects.size(); i++) {
	    add(newObjects.get(i), index + i);
	}
    }

    /**
     * Remove the highest indexed object of this array list. If after
     * removing the object the number of objects stored in the array
     * is less than one quarter of the array length and the array
     * length is greater than its initial length, then the array is
     * halved in order to save memory.
     * 
     * @param index to add list of objects
     * @throws IndexOutOfBoundsException if the list is empty
     * @return the object that was removed
     */
    public final T remove() {
        return remove(size - 1);
    }

    /**
     * Remove the object at the specified index of this array
     * list. All objects that have a higher index are shifted down one
     * index in the array to fill the gap created by removing the
     * object. If after removing the object the number of objects
     * stored in the array is less than one quarter of the array
     * length and the array length is greater than its initial length,
     * then the array is halved in order to save memory.
     * 
     * @param index the index of the object to remove.
     * @throws IndexOutOfBoundsException if index < 0 or index >= size
     * @return the object that was removed
     */
    public final T remove(final int index) {
	if (index < 0 || index > size) {
	    throw new IndexOutOfBoundsException("Invalid index: " + index);
	}
	T removedObj = objects[index];
	for (int i = index; i < size - 1; i++) {
	    objects[i] = objects[i + 1];
	}
	size--;
	if (size <= objects.length / 4) {
	    halveCapacity();
	}
	return removedObj;
    }

    /**
     * Remove objects between and including the specified begin index
     * and the end of the list. If after removing the objects the
     * number of objects stored in the array is less than one quarter
     * of the array length and the array length is greater than its
     * initial length, then the array is halved in order to save
     * memory. If the begin index is not valid the list is unchanged
     * and null is returned.
     * 
     * @param beginIndex index of first object to remove
     * @throws IndexOutOfBoundsException if beginIndex < 0 or beginIndex >= size
     * @return the list of objects that were removed
     */
    public final ArrayList<T> removeList(final int beginIndex) {
        return removeList(beginIndex, size);
    }

    /**
     * Remove objects between and including the specified begin index
     * and excluding the specified end index. All objects that have a
     * higher index than the highest indexed object that is removed
     * are shifted down by the number of objects that are removed to
     * fill the gap created by removing the objects. If after removing
     * the objects the number of objects stored in the array is less
     * than one quarter of the array length and the array length is
     * greater than its initial length, then the array is halved in
     * order to save memory.
     * 
     * @param beginIndex index of first object to remove
     * @param endIndex index of the last object to remove (exclusive)
     * @throws IndexOutOfBoundsException if either beginIndex < 0,
     * endIndex > size, or beginIndex >= endIndex
     * @return the list of objects that were removed
     */
    public final ArrayList<T> removeList(final int beginIndex,
					 final int endIndex) {
	if (beginIndex < 0 || endIndex > size || beginIndex >= endIndex) {
	    throw new IndexOutOfBoundsException("Invalid indices: " + beginIndex
					       + ", " + endIndex);
	}
	ArrayList<T> removedList = new ArrayList<T>();
	int numToRemove = endIndex - beginIndex;
	for (int i = 0; i < numToRemove; i++) {
	    removedList.add(remove(beginIndex));
	}
	return removedList;
    }

    /**
     * Doubles the capacity of the objects array without changing any
     * of the contents.
     */
    @SuppressWarnings("unchecked")
    private final void doubleCapacity() {
	T[] tempObjects = (T[]) new Object[objects.length * 2];
	for (int objIter = 0; objIter < objects.length; objIter++) {
	    tempObjects[objIter] = objects[objIter];
	}
	objects = tempObjects;
    }

    /**
     * Halves the capacity of the objects array without changing any
     * of the contents.  If the new capacity is less than the current
     * size then elements at indices greater than size - 1 will be
     * removed.
     */
    @SuppressWarnings("unchecked")
    private final void halveCapacity() {
	T[] tempObjects = (T[]) new Object[objects.length / 2];
	if (tempObjects.length < size) {
	    size = tempObjects.length;
	}
	for (int objIter = 0; objIter < size; objIter++) {
	    tempObjects[objIter] = objects[objIter];
	}
	objects = tempObjects;
    }

    /**
     * Randomly reorders the objects in this list by repeatedly
     * choosing two indices and swapping the two objects.
     */
    public final void shuffle() {
        Random rand = new Random();
        for (int objIter = 0; objIter < size - 1; objIter++) {
            swap(objIter, rand.nextInt(size - objIter) + objIter);
        }
    }

    /**
     * Sorts this list from smallest to largest using the bubble sort
     * algorithm.
     *
     * @throws ClassCastException if the type of the obects in the
     * array do not implement Comparable
     */
    public final void bubbleSort() {
        for (int iter = size; iter > 0; iter--) {
	    bubble(iter);
        }

    }

    /**
     * Bubble up the largest element between the first index and the
     * specified index until it reaches the specified index.
     * 
     * @throws ClassCastException if the type of the obects in the
     * array do not implement Comparable
     */
    private final void bubble(final int index) {
	for (int iter = 0; iter < index - 1; iter++) {
	    if (compare(objects[iter], objects[iter + 1]) > 0) {
		swap(iter, iter + 1);
	    }
	}
    }

    /**
     * Sorts this list from smallest to largest using the merge sort
     * algorithm.
     *
     * @throws ClassCastException if the type of the obects in the
     * array do not implement Comparable
     */
    public final void mergeSort() {
	mergeSort(0, size);
    }

    /**
     * Sorts the sub-array between the specified indices using the
     * bubble sort algorithm.
     *
     * @param begin index of the first element of the sub-array, inclusive
     * @param end index of the last element of the sub-array, exclusive
     * @throws IndexOutOfBoundsException if begin < 0, end > size, or
     * begin >= end
     * @throws ClassCastException if the type of the obects in the
     * array do not implement Comparable
     */
    private final void mergeSort(final int beginIndex, final int endIndex) {
	if (endIndex - beginIndex > 1) {
            int midIndex = (beginIndex + endIndex) / 2;
            mergeSort(beginIndex, midIndex);
            mergeSort(midIndex, endIndex);
            merge(beginIndex, midIndex, endIndex);
        }
    }

    /**
     * Merge, or reorder, the elements between the begin and end
     * indices such that the element are in order from smallest to
     * largest.  The sub-array between the indices begin and middle
     * must be ordered from smallest to largest and the sub-array
     * between the indices middle and end must be ordered from
     * smallest to largest.
     *
     * @param begin index of the first element of the right sub-array, inclusive
     * @param middle index of the last element of the left sub-array,
     * inclusive, and the index of the first element of the right
     * sub-array, exclusive
     * @param end index of the last element of the left sub-array, exclusive
     * @throws IndexOutOfBoundsException if begin < 0, end > size,
     * middle >= end, or begin >= middle
     * @throws ClassCastException if the type of the obects in the
     * array do not implement Comparable
     */
    public final void merge(final int beginIndex, final int middleIndex, 
			    final int endIndex) {
	int tempSize = endIndex - beginIndex; // size of temp array
	T[] temp = (T[]) new Object[tempSize]; // merged sub array
        int left = beginIndex; // left sub-array iterator
        int right = middleIndex; // right sub-array iterator
        int index = 0; // temp array iterator

	/* Merge the left and right sub-arrays until end of one is reached. */
        while (left < middleIndex && right < endIndex) {
	    if (compare(objects[left], objects[right]) < 0) {
		temp[index] = objects[left];
		left++;
	    } else {
		temp[index] = objects[right];
		right++;
	    }
            index++;
        }

	/* Copy the rest of left sub-array, if anything remains. */
        while (left < middleIndex) {
	    temp[index] = objects[left];
            left++;
	    index++;
        }

	/* Copy the rest of right sub-array, if anything remains. */
        while (right < endIndex) {
	    temp[index] = objects[right];
            right++;
	    index++;
        }

	/* Copy sorted temp sub-array back to objects array. */
        for (int iter = 0; iter < temp.length; iter++) {
	    objects[iter + beginIndex] = temp[iter];
        }
    }

    /**
     * Sorts this list from smallest to largest using the selection sort
     * algorithm.
     *
     * @throws ClassCastException if the type of the obects in the
     * array do not implement Comparable
     */
    public final void selectionSort() {
	for (int i = 0; i < size; i++) {
	    swap(i, indexOfSmallest(i, size));
	}
    }

    /**
     * If the objects in this list implement comparable, the index of
     * the smallest object in the array between the specified indices
     * is returned.  If there are multiple smallest objects in the
     * array, the smallest index of all the smallest objects is
     * returned.  If the objects are not comparable or if the indices
     * are not valid, -1 is returned.
     *
     * @param begin index of first element in sub array to search
     * @param end index of last element in sub array to search (exclusive)
     * @throws IndexOutOfBoundsException if begin < 0, end > size, or
     * begin >= end
     * @throws ClassCastException if the type of the obects in the
     * array do not implement Comparable
     * @return index of smallest object
     */
    private final int indexOfSmallest(int begin, int end) {
	if (begin < 0 || end > size || begin >= end) {
            throw new IndexOutOfBoundsException("Invalid indices: " + begin +
					       ", " + end);
	}
	T minObj = objects[begin];
	int minIndex = begin;
	for (int objIter = begin + 1; objIter < end; objIter++) {
	    if (compare(objects[objIter], minObj) < 0) {
		minObj = objects[objIter];
		minIndex = objIter;
	    }
	}
	return minIndex;
    }

    /**
     * Sorts this list from smallest to largest using the insertion sort
     * algorithm.
     *
     * @throws ClassCastException if the type of the obects in the
     * array do not implement Comparable
     */
    public final void insertionSort() {
	for (int iter = 1; iter < size(); iter++) {
	    insertObjectAtIndex(iter);
	}
    }

    /**
     * If the elements of this before the specified index are in
     * ascending order, the element at the specified index is inserted
     * into the correct location such that the elements from the
     * beginning to the index are sorted.
     *
     * @param index the index of the element to insert
     * @throws ClassCastException if the type of the obects in the
     * array do not implement Comparable
     * @throws IndexOutOfBoundsException if index < 0 or index >= size
     */
    private final void insertObjectAtIndex(final int index) {
	if (index < 0 || index >= size) {
	    throw new IndexOutOfBoundsException("invalid index: " + index);
	}
	T insertObj = objects[index];
	int iter = index;
	while (iter > 0 && compare(objects[iter - 1], insertObj) > 0) {
	    objects[iter] = objects[iter - 1];
	    iter--;
	}
	objects[iter] = insertObj;
    }

    /**
     * Returns 0 if the two objects are equal, negative integer if
     * left is less than right, or positive integer if left is greater
     * than right.  Null objects are considered to be larger than any
     * non-null object.
     *
     * @param left object to compare
     * @param right object to compare
     * @throws ClassCastException if the type of the obects in the
     * array do not implement Comparable
     */
    @SuppressWarnings("unchecked")
    private final int compare(final T left, final T right) {
	if (left == null && right == null) {
	    return 0;
	}
	if (left != null && right == null) {
	    return -1;
	}
	if (left == null && right != null) {
	    return 1;
	}
	return ((Comparable<T>) left).compareTo(right);
    }

    /**
     * Swap the two objects at the specified indices. If the indices
     * are invalid the list is unchanged.
     * 
     * @param index1 index of the first object
     * @param index2 index of the second object
     * @throws IndexOutOfBoundsException if either index is out of range
     */
    private final void swap(final int index1, final int index2) {
        if (index1 < 0 || index1 >= size || index2 < 0 || index2 >= size) {
            throw new IndexOutOfBoundsException("Invalid indices: " + index1 +
					       ", " + index2);
	}
	if (index1 == index2) {
	    return;
	}
	T tmpObj = objects[index2];
	objects[index2] = objects[index1];
	objects[index1] = tmpObj;
    }

    /**
     * Returns the object in the list at the specified index.
     * 
     * @param index index of element to get
     * @throws IndexOutOfBoundsException if index < 0 or index >= size
     * @return object at specified index
     */
    public final T get(final int index) {
	if (index < 0 || index >= size) {
	    throw new IndexOutOfBoundsException("Invalid index: " + index);
	}
        return objects[index];
    }

    /**
     * Returns the first object in the list.
     * 
     * @throws IndexOutOfBoundsException if the list is empty
     * @return the first object in the list
     */
    public final T getFirst() {
	return get(0);
    }

    /**
     * Returns the last object in the list.
     * 
     * @throws IndexOutOfBoundsException if the list is empty
     * @return the last object in the list
     */
    public final T getLast() {
	return get(size - 1);
    }

    /**
     * Sets the object in this list to be the specified object.
     * 
     * @param object object to set
     * @param index index to set object
     * @throw IndexOutOfBoundsException if index < 0 or index >= size
     */
    public final void set(final T object, final int index) {
        if (index < 0 || index >= size) {
            throw new IndexOutOfBoundsException("Invalid index: " + index);
        }
        objects[index] = object;
    }

    /**
     * Returns a copy of this list without copying any of the objects in this
     * list.
     * 
     * @return a shallow copy of this list
     */
    @SuppressWarnings("unchecked")
    public final ArrayList<T> clone() {
        ArrayList<T> newList = new ArrayList<T>();
        newList.objects = (T[]) new Object[objects.length];
        System.arraycopy(objects, 0, newList.objects, 0, size);
        newList.size = size;
        return newList;
    }

    /**
     * Returns the number of objects currently stored in this array
     * list.
     * 
     * @return the number of objects in this array list
     */
    public final int size() {
        return size;
    }

    /**
     * Returns a string representation of this array list.
     * 
     * @return a string representation of this array list
     */
    public final String toString() {
	String string = "[";
	for (int objectIter = 0; objectIter < size - 1; objectIter++) {
	    string += objects[objectIter] + ", ";
	}
	if (size > 0) {
	    string += objects[size - 1];
	}
	string += "]";
        return string;
    }
}
